﻿#include "stdafx.h"
#include "UnitTestFrameWork.h"
#include "TestP4BridgeServer.h"
#include "TextEncoder.h"

#include "../p4bridge/P4BridgeClient.h"
#include "../p4bridge/P4BridgeServer.h"
#include "../p4bridge/P4BridgeEnviro.h"
#include "../p4bridge/P4Connection.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <conio.h>
#include <ios>
#include <string>
#include <sstream>
#include <fstream>

// Does not return the right value, but is available in VS2010, and does most of what we want
#define snprintf(buf,len, format,...) _snprintf_s(buf, len,len, format, __VA_ARGS__)

CREATE_TEST_SUITE(TestP4BridgeServer)

TestP4BridgeServer::TestP4BridgeServer(void)
{
    UnitTestSuite::RegisterTest(ServerConnectionTest, "ServerConnectionTest");
    UnitTestSuite::RegisterTest(TestUnicodeClientToNonUnicodeServer, "TestUnicodeClientToNonUnicodeServer");
    UnitTestSuite::RegisterTest(TestUnicodeUserName, "TestUnicodeUserName");
    UnitTestSuite::RegisterTest(TestUntaggedCommand, "TestUntaggedCommand");
    UnitTestSuite::RegisterTest(TestTaggedCommand, "TestTaggedCommand");
    UnitTestSuite::RegisterTest(TestTextOutCommand, "TestTextOutCommand");
    UnitTestSuite::RegisterTest(TestBinaryOutCommand, "TestBinaryOutCommand");
    UnitTestSuite::RegisterTest(TestErrorOutCommand, "TestErrorOutCommand");
    UnitTestSuite::RegisterTest(TestGetSet, "TestGetSet");
    UnitTestSuite::RegisterTest(TestEnviro, "TestEnviro");
    UnitTestSuite::RegisterTest(TestGetConfig, "TestGetConfig");
    UnitTestSuite::RegisterTest(TestIsIgnored, "TestIsIgnored");
    UnitTestSuite::RegisterTest(TestConnectionManager, "TestConnectionManager");
    UnitTestSuite::RegisterTest(TestSetVars, "TestSetVars");
    UnitTestSuite::RegisterTest(TestParallelSync, "TestParallelSync"); 
    UnitTestSuite::RegisterTest(TestDefaultProgramNameAndVersion, "TestDefaultProgramNameAndVersion");
    UnitTestSuite::RegisterTest(TestGetTicketFile, "TestGetTicketFile");
}


TestP4BridgeServer::~TestP4BridgeServer(void)
{
}

char unitTestDir[MAX_PATH];
char unitTestZip[MAX_PATH];
char * TestDir = "c:\\MyTestDir";
char * TestZip = "c:\\MyTestDir\\a.exe";
char * rcp_cmd = "p4d -r C:/MyTestDir -jr checkpoint.1";
char * udb_cmd = "p4d -r C:/MyTestDir -xu";
char * p4d_cmd = "p4d -p6666 -IdUnitTestServer -rC:/MyTestDir -Llog";
char * testCfgFile = "C:\\MyTestDir\\admin_space\\myP4Config.txt";
char * testCfgDir = "C:\\MyTestDir\\admin_space";
char * testIgnoreFile = "C:\\MyTestDir\\admin_space\\myP4Ignore.txt";
char * testIgnoredFile1 = "C:\\MyTestDir\\admin_space\\foofoofoo.foo";
char * testIgnoredFile2 = "C:\\MyTestDir\\admin_space\\moomoomoo.moo";
char * testProgramName = "BridgeUnitTests";
char * testProgramVer = "1.2.3.4.A.b.C";

void * pi = NULL;
P4BridgeServer * ps = NULL;

char* oldConfig = NULL;
char* oldIgnore = NULL;

bool TestP4BridgeServer::Setup()
{
    // remove the test directory if it exists
    UnitTestSuite::rmDir( TestDir ) ;

    GetCurrentDirectory(sizeof(unitTestDir), unitTestDir);

    strcpy( unitTestZip, unitTestDir);
    strcat( unitTestZip, "\\a.exe");

    if (!CreateDirectory( TestDir, NULL))
    {
		DWORD err = GetLastError();
		printf("GetLastError returns %ld\n",err);

	    return false;
    }
		
    if (!CopyFile(unitTestZip, TestZip, false)) 
		return false;

    if (!SetCurrentDirectory(TestDir)) 
		return false;

    pi = UnitTestSuite::RunProgram("a", TestDir, true, true);
    if (!pi) 
    {
        SetCurrentDirectory(unitTestDir);
        return false;
    }

    delete pi;

    pi = UnitTestSuite::RunProgram(rcp_cmd, TestDir, true, true);
    if (!pi) 
    {
        SetCurrentDirectory(unitTestDir);
        return false;
    }

    delete pi;

    pi = UnitTestSuite::RunProgram(udb_cmd, TestDir, true, true);
    if (!pi) 
    {
        SetCurrentDirectory(unitTestDir);
        return false;
    }

    delete pi;

    pi = UnitTestSuite::RunProgram(p4d_cmd, TestDir, false, false);
    if (!pi) 
    {
        SetCurrentDirectory(unitTestDir);
        return false;
    }

//    _getch();

    return true;
}

int TestP4BridgeServer::LogCallback(int level, const char *file, int line, const char *msg)
{
	printf("LOG: %d %s:%d %s", level, file, line , msg);
	return 1;
}

bool TestP4BridgeServer::TearDown(char* testName)
{
	UnitTestSuite::rmDir( testCfgFile ) ;
    UnitTestSuite::rmDir( testIgnoreFile ) ;
	if (oldConfig)
	{
		P4BridgeServer::Set("P4CONFIG", oldConfig);	
		oldConfig = NULL;
	}
	if (oldIgnore)
	{
		P4BridgeServer::Set("P4IGNORE", oldIgnore);	
		oldIgnore = NULL;
	}
    if (pi)
        UnitTestSuite::EndProcess( (LPPROCESS_INFORMATION) pi );
	if (ps)
    {
		delete ps;
		ps = NULL;

		//p4base::DumpMemoryState("After deleting server");
	}
    SetCurrentDirectory(unitTestDir);

    UnitTestSuite::rmDir( TestDir ) ;

	p4base::PrintMemoryState(testName);

    return true;
}

bool TestP4BridgeServer::ServerConnectionTest()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "");
    ASSERT_NOT_NULL(ps);
	//ps->SetLogCallFn((LogCallbackFn *) &TestP4BridgeServer::LogCallback);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }

    return true;
}

bool TestP4BridgeServer::TestUnicodeClientToNonUnicodeServer()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }

    ASSERT_NOT_EQUAL(ps->unicodeServer(), 1);
    ps->set_charset("utf8", "utf16le");

    char* params[1];
    params[0] = "//depot/mycode/*";

    ASSERT_FALSE(ps->run_command("files", 3456, 0, params, 1))

    P4ClientError * out = ps->get_ui(3456)->GetErrorResults();

    ASSERT_STRING_STARTS_WITH(out->Message, "Unicode clients require a unicode enabled server.")

	ps->ReleaseConnection(3456,0);

    return true;
}

bool TestP4BridgeServer::TestUnicodeUserName()
{

    P4ClientError* connectionError = NULL;
    // create a new server
    //Aleksey (Alexei) in Cyrillic = "\xD0\x90\xD0\xbb\xD0\xB5\xD0\xBA\xD1\x81\xD0\xB5\xD0\xB9\0" IN utf-8
    ps = new P4BridgeServer("localhost:6666", "\xD0\x90\xD0\xBB\xD0\xB5\xD0\xBA\xD1\x81\xD0\xB5\xD0\xB9\0", "pass", "\xD0\x90\xD0\xbb\xD0\xB5\xD0\xBA\xD1\x81\xD0\xB5\xD0\xB9\0");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }

    ASSERT_FALSE(ps->unicodeServer());
    ps->set_charset("utf8", "utf16le");
	// ps->set_charset("none", "none");                                  // This disables client P4CHARSET

    char* params[1];
    params[0] = "//depot/mycode/*";

    ASSERT_FALSE(ps->run_command("files", 7, 0, params, 1))

    P4ClientError * out = ps->get_ui(7)->GetErrorResults();

    ASSERT_STRING_STARTS_WITH(out->Message, "Unicode clients require a unicode enabled server.")

    return true;
}

bool TestP4BridgeServer::TestUntaggedCommand()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", *connectionError);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/mycode/*";

    ASSERT_INT_TRUE(ps->run_command("files", 7, 0, params, 1))

    P4ClientInfoMsg * out = ps->get_ui(7)->GetInfoResults();

    ASSERT_STRING_STARTS_WITH(out->Message, "//depot/MyCode/")

    return true;
}

bool TestP4BridgeServer::TestTaggedCommand()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/mycode/*";

    ASSERT_INT_TRUE(ps->run_command("files", 7, 1, params, 1))

    StrDictListIterator * out = ps->get_ui(7)->GetTaggedOutput();

    ASSERT_NOT_NULL(out);

    int itemCnt = 0;
    while (StrDictList * pItem = out->GetNextItem())
    {
        int entryCnt = 0;

        while (KeyValuePair * pEntry = out->GetNextEntry())
        {
            if ((itemCnt == 0) && (strcmp(pEntry->key, "depotFile") == 0))
                ASSERT_STRING_STARTS_WITH(pEntry->value, "//depot/MyCode/")
            if ((itemCnt == 1) && (strcmp(pEntry->key, "depotFile") == 0))
                ASSERT_STRING_STARTS_WITH(pEntry->value, "//depot/MyCode/")
            entryCnt++;
        }
        ASSERT_NOT_EQUAL(entryCnt, 0);
        itemCnt++;
    }
    ASSERT_EQUAL(itemCnt, 3);

	delete out;

    return true;
}

bool TestP4BridgeServer::TestTextOutCommand()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/MyCode/ReadMe.txt";

    ASSERT_INT_TRUE(ps->run_command("print", 7, 1, params, 1))

    StrBuf * out = ps->get_ui(7)->GetTextResults();

    ASSERT_NOT_NULL(out);

    ASSERT_STRING_EQUAL(out->Text(), "Don't Read This!\n\nIt's Secret!")

    return true;
}

bool TestP4BridgeServer::TestBinaryOutCommand()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/MyCode/Silly.bmp";

    ASSERT_INT_TRUE(ps->run_command("print", 7, 1, params, 1))

    int cnt = ps->get_ui(7)->GetBinaryResultsCount();

    ASSERT_EQUAL(cnt, 3126)

    void * out = ps->get_ui(7)->GetBinaryResults();

    ASSERT_NOT_NULL(out);
    ASSERT_EQUAL((*(((unsigned char*)out) + 1)), 0x4d)

    return true;
}

bool TestP4BridgeServer::TestErrorOutCommand()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255,"Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/MyCode/Billy.bmp";

    // run a command against a nonexistent file
    // Should fail
    ASSERT_FALSE(ps->run_command("rent", 7, 1, params, 1))

    P4ClientError * out = ps->get_ui(7)->GetErrorResults();

    ASSERT_NOT_NULL(out);

    //ASSERT_STRING_STARTS_WITH(out->Message, "Unknown command.  Try 'p4 help' for info")
	ASSERT_EQUAL(out->ErrorCode, 805379098);
    ASSERT_NULL(out->Next)

    return true;
}

bool TestP4BridgeServer::TestGetSet()
{
	char *expected = "C:\\login.bat";

	char *test_charset = "utf16";

	P4ClientError* connectionError = NULL;
	// create a new server
	ps = new P4BridgeServer("localhost:6666", "admin", "", "");
	ASSERT_NOT_NULL(ps);
	ps->SetLogCallFn((LogCallbackFn *) &TestP4BridgeServer::LogCallback);
	ASSERT_NOT_NULL(ps);

	// set server to use test_charset for file translation
	ps->set_charset(test_charset);

	// get charset from server
	char *cset = ps->get_charset()->Text();

	ASSERT_STRING_EQUAL(test_charset, cset);

	// Test the P4BridgeEnviro Update() function
	// Any values set with Update should be the first returned.
	P4BridgeServer::Update("P4FOOBAR", expected);

	char *result = P4BridgeServer::Get("P4FOOBAR");

	ASSERT_STRING_EQUAL(expected, result);
	delete[] result;

	P4BridgeServer::ClearUpdate("P4FOOBAR");
	result = P4BridgeServer::Get("P4FOOBAR");
	ASSERT_NULL(result);

	wchar_t *w_expected = L"C:\\login<АБВГ>.bat";
	expected = TextEncoder::Utf16ToUtf8(w_expected);

	P4BridgeServer::Set("P4FOOBAR", expected);

	result = P4BridgeServer::Get("P4FOOBAR");

	ASSERT_STRING_EQUAL(expected, result);

	wchar_t *w_result = TextEncoder::Utf8ToUtf16(result);

	ASSERT_W_STRING_EQUAL(w_expected, w_result);

	delete[] result;

	P4BridgeServer::Set("P4FOOBAR", NULL);
	result = P4BridgeServer::Get("P4FOOBAR");
	ASSERT_NULL(result);

   return true;
}

/* TestEnviro utility routines */

char * get_value(char *var)
{
	P4BridgeServer::enviro.SetCharSet(1); // set utf 8

	char* copy = CopyStr(P4BridgeServer::enviro.Get( var ));
	return copy;
}

void update_value(char *var, char *value)
{
	P4BridgeServer::enviro.SetCharSet(1); // set utf 8

	P4BridgeServer::enviro.Update(var,value);
}

void clear_value(char *var)
{
	P4BridgeServer::enviro.SetCharSet(1); // set utf 8

	P4BridgeServer::enviro.ClearUpdate(var);
}

bool TestP4BridgeServer::TestEnviro()
{
	// Save the existing value
	char *existing = get_value("P4CONFIG");

	// Override the value with Update()
	update_value("P4CONFIG", "myP4Config.txt");
	char *result1 = get_value("P4CONFIG");
	printf("result1: %s\n",result1);

	// Check with a value not in the environment
	update_value("P4NOENV", "myP4NOENV.txt");
	char *result2 = get_value("P4NOENV");
	printf("result2: %s\n",result2);

	ASSERT_NOT_NULL(result2);
	ASSERT_STRING_EQUAL(result2, "myP4NOENV.txt");

	// Now try to clear the value locally
	update_value("P4CONFIG","");
	char *result3 = get_value("P4CONFIG");
	ASSERT_STRING_EQUAL(result3,"");

	// Now remove the Local override
	clear_value("P4CONFIG");
	char *result4 = get_value("P4CONFIG");

	ASSERT_STRING_EQUAL(result4, existing);
	
    return true;
}

bool TestP4BridgeServer::TestGetConfig()
{
	OFSTRUCT ReOpenBuff;

	P4BridgeServer::ClearUpdate("P4CONFIG");	// remove any Local Value for P4CONFIG

	oldConfig = P4BridgeServer::Get("P4CONFIG");

	HFILE hf = OpenFile(testCfgFile, &ReOpenBuff, OF_CREATE);
	DWORD written = 0;
	WriteFile((HANDLE) hf, (LPVOID) "P4CLIENT testClient\r\n", 21, &written, NULL);
	CloseHandle((HANDLE) hf);

	P4BridgeServer::Update("P4CONFIG", "myP4Config.txt");	

	char *cfg = P4BridgeServer::Get("P4CONFIG"); 
	ASSERT_STRING_EQUAL(cfg, "myP4Config.txt");

	char *result = P4BridgeServer::get_config(testCfgDir);

	ASSERT_STRING_EQUAL(testCfgFile, result);

	delete[] result;

    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

	ps->set_cwd(testCfgDir);

	result = P4BridgeServer::Get("P4CONFIG");
	
	ASSERT_STRING_EQUAL("myP4Config.txt", result);

	result = ps->get_config();

	ASSERT_STRING_EQUAL(testCfgFile, result);

    return true;
}

bool TestP4BridgeServer::TestIsIgnored()
{
	OFSTRUCT ReOpenBuff;

	oldIgnore = P4BridgeServer::Get("P4IGNORE");

	HFILE hf = OpenFile(testIgnoreFile, &ReOpenBuff, OF_CREATE);
	DWORD written = 0;
	WriteFile((HANDLE) hf, (LPVOID) "foofoofoo.foo\r\n", 15, &written, NULL);
	CloseHandle((HANDLE) hf);

	P4BridgeServer::Set("P4IGNORE", "myP4Ignore.txt");	

	ASSERT_INT_TRUE(P4BridgeServer::IsIgnored(StrRef(testIgnoredFile1)));
	ASSERT_FALSE(P4BridgeServer::IsIgnored(StrRef(testIgnoredFile2)));

	return true;
}

bool TestP4BridgeServer::TestConnectionManager()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/mycode/*";

	//p4base::DumpMemoryState("Before first command run");

	ASSERT_INT_TRUE(ps->run_command("files", 6, 1, params, 1))

	//p4base::DumpMemoryState("After first command run");

	ps->ReleaseConnection(6, 10);

	//p4base::DumpMemoryState("After first ReleaseConnection");

    ASSERT_INT_TRUE(ps->run_command("files", 7, 1, params, 1))

	//p4base::DumpMemoryState("After second command run");

	ps->ReleaseConnection(7, 10);

	//p4base::DumpMemoryState("After second ReleaseConnection");

	int idleCnt = ps->FreeConnections(0);

    ASSERT_EQUAL(idleCnt, 1)

	//p4base::DumpMemoryState("After first FreeConnections");

    ASSERT_INT_TRUE(ps->run_command("files", 8, 1, params, 1))	

	//p4base::DumpMemoryState("After third command run");

    ASSERT_INT_TRUE(ps->run_command("files", 9, 1, params, 1))

	//p4base::DumpMemoryState("After fourth command run");

	ps->ReleaseConnection(8, 20);
	//p4base::DumpMemoryState("After third ReleaseConnection");
	ps->ReleaseConnection(9, 30);
	//p4base::DumpMemoryState("After fourth ReleaseConnection");

	idleCnt = ps->FreeConnections(11);
    ASSERT_EQUAL(idleCnt, 2)
	//p4base::DumpMemoryState("After second FreeConnections");

	idleCnt = ps->FreeConnections(21);
    ASSERT_EQUAL(idleCnt, 1)
	//p4base::DumpMemoryState("After third FreeConnections");

	idleCnt = ps->FreeConnections(31);
    ASSERT_EQUAL(idleCnt, 0)
	//p4base::DumpMemoryState("After fourth FreeConnections");

    return true;
}

//    Some experimentation with command options happened in this test.
//     There is a lot of supporting code which is still there, but unneeded for just the test
bool TestP4BridgeServer::TestSetVars()
{
	Error e;

	// test data for reading Protocols (these are read after client::Init())
	char *name = "apilevel";
	int value = 38;

	char *name1 = "server";
	char *value1 = "40";

	char *name2 = "nocase";
	char *value2 = "1";

	char *name3 = "security";
	char *value3 = "0";

	char *name4 = "unicode";
	char *value4 = "0";
 
	P4ClientError* connectionError = NULL;
	StrPtr *pptr = NULL;

	// create a new server
	ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
	ASSERT_NOT_NULL(ps);

	// enable logging of C++ events
	//ps->SetLogCallFn((LogCallbackFn *)&TestP4BridgeServer::LogCallback);  // Works
		
	// Protocol options must be set before client::Init()
	// -Z options are set by Protocol
	//ps->SetProtocol("dbstat","");    // "track", "dbstat", "app=name", "proxyverbose"
	//ps->SetProtocol(P4Tag::v_tag, "yes");  // enable tagged output  WORKS
	// ps->SetProtocol("track", ""); // enable tracking WORKS

	// connect and see if the api returned an error. 
	if (!ps->connected(&connectionError))
	{
		char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
		// Abort if the connect did not succeed
		ASSERT_FAIL(buff);
	}

	P4Connection* client = ps->ConMgr->GetConnection(0);
	P4BridgeClient *bridge_ui = ps->get_ui(0);

	// What command to send?
	char * cmdname = "users";
	char * const v[2] = {"-m", "10"};
	client->SetArgv(2, v);

	// Interestingly,  the "tag" option works as both a Var and a Protocol
	//client->SetVar(P4Tag::v_tag, "yes");  // enable tagged output  WORKS

	// -z options maxLockTime maxOpenFiles maxResults maxScanRows
	//client->SetVarV("maxOpenFiles=1");  

	// -v options are passed by p4debug.SetLevel() calls.
	// p4debug.SetLevel("rpc=3");   // This will spew debug output to stdout WORKS
	// p4debug.SetLevel(5);		// This will spew ALL debug output to stdout WORKS

	//p4debug.SetLevel("track=1");

	ps->Run_int(client, cmdname, bridge_ui);

	//p4debug.SetLevel(0);   // Disable Debug output WORKS

	StrDictListIterator * tagout = bridge_ui->GetTaggedOutput();	
	StrBuf * tout = bridge_ui->GetTextResults();	
	StrPtr * tcount = bridge_ui->GetDataSet();

	void * bout = bridge_ui->GetBinaryResults();
	int bcount = bridge_ui->GetBinaryResultsCount();

	P4ClientInfoMsg *iout = bridge_ui->GetInfoResults();
	int icount = bridge_ui->GetInfoResultsCount();
	P4ClientError *eout = bridge_ui->GetErrorResults();

	if (icount > 0 && iout)
	{
		P4ClientInfoMsg *imsg = iout;
		while (imsg)
		{
			printf("info: %d %s\n", imsg->Level, imsg->Message);
			imsg = imsg->Next;
		}
			
	}
	//ASSERT_NOT_NULL(tagout);

	if (tagout)  // Tagged output
	{
		int itemCnt = 0;
		while (StrDictList * pItem = tagout->GetNextItem())
		{
			int entryCnt = 0;

			while (KeyValuePair * pEntry = tagout->GetNextEntry())
			{
				printf("tag: %d %s => %s\n", itemCnt, pEntry->key, pEntry->value);
				entryCnt++;
			}
			itemCnt++;
		}
	}

	if (tout) // Text output
	{
		printf("text: %s", tout->Value());
	}


	int level = ps->APILevel();

	if (level < value)   // Server2 38 or larger is ok
	{
		char buff[256];
		snprintf(buff, 255, "API mismatch %d < %d", level, value);
		ASSERT_FAIL(buff);
	}

	pptr = client->GetProtocol(name1);
	ASSERT_NOT_NULL(pptr);

	if (strcmp(value1, pptr->Value()))
	{
		char buff[256];
		snprintf(buff, 255, "Value1 mismatch %s != %s", value1, pptr->Value());
		ASSERT_FAIL(buff);
	}

	pptr = client->GetProtocol(name2);
	ASSERT_NOT_NULL(pptr);

	if (strcmp(value2, pptr->Value()))
	{
		char buff[256];
		snprintf(buff, 255, "Value2 mismatch %s != %s", value2, pptr->Value());
		ASSERT_FAIL(buff);
	}

	pptr = client->GetProtocol(name3);
	ASSERT_NOT_NULL(pptr);

	if (strcmp(value3, pptr->Value()))
	{
		char buff[256];
		snprintf(buff, 255, "Value3 mismatch %s != %s", value3, pptr->Value());
		ASSERT_FAIL(buff);
	}

	pptr = client->GetProtocol(name4);
	ASSERT_NOT_NULL(pptr);

	if (strcmp(value4, pptr->Value()))
	{
		char buff[256];
		snprintf(buff, 255, "Value4 mismatch %s != %s", value4, pptr->Value());
		ASSERT_FAIL(buff);
	}
	return true;
}

// A test for Parallel Sync error handling - Job085941

bool TestP4BridgeServer::TestParallelSync()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }

    ASSERT_NOT_EQUAL(ps->unicodeServer(), 1);
   // ps->set_charset("utf8", "utf16le");

	char *args[4];

	// run p4 configure set net.parallel.max=4	
	args[0] = "set";
	args[1] = "net.parallel.max=4";
	ASSERT_INT_TRUE(ps->run_command("configure", 0, 1, args, 2));

	// run p4 configure set dmc=3
	args[0] = "set";
	args[1] = "dmc=3";
	ASSERT_INT_TRUE(ps->run_command("configure", 0, 1, args, 2));

	// Paths of interest for the parallel test.
	std::string ppath = "c:\\MyTestDir\\admin_space\\TestData\\parallel\\";
	std::string ppathall = ppath + "...";
	std::string ppathnone = ppathall + "#none";
	std::string targetpath = ppath + "testFile99.txt";

	std::string ppath1 = "c:\\MyTestDir\\admin_space2\\TestData\\parallel\\";
	std::string targetpath1 = ppath1 + "testFile99.txt";

	// Create the workspace directory for the parallel test
	CreateDirectory(ppath.c_str(), NULL);
	
	// write 500 files of 1K size to the workspace directory
	int count = 500;
	for (int i = 0; i < count; i++)
	{
		std::ostringstream oss;
		oss << ppath << "testFile" << i << ".txt";

		std::ofstream ofs(oss.str(),std::ios::binary | std::ios::out);
		ofs.seekp(1020);
		ofs.write("test", 4);
		ofs.close();
	}

	// add them to the server
	args[0] = const_cast<char *>(ppathall.c_str());
	ASSERT_INT_TRUE(ps->run_command("add",0,1,args,1));

	// submit them
	args[0] = "-d";
	args[1] = "\"initial submit of test files\"";
	args[2] = const_cast<char *>(ppathall.c_str());
	ASSERT_INT_TRUE(ps->run_command("submit",0,1,args,3));

	// Remove all Workspace Files (Sync to NONE)
	args[0] = "-f";
	args[1] = const_cast<char *>(ppathnone.c_str());
	ASSERT_INT_TRUE(ps->run_command("sync",0,1,args,2));

	// Sync just the target file
	args[0] = "-f";
	args[1] = const_cast<char *>(targetpath.c_str());
	ASSERT_INT_TRUE(ps->run_command("sync",0,1,args,2));

	// Change target file to Writable
	_chmod(targetpath.c_str(), _S_IWRITE | _S_IREAD);

	// Save current user and client
	//const StrPtr *user1 = ps->get_user();
	//const StrPtr *client1 = ps->get_client();

	// Change to a different client
	ps->set_client("admin_space2");

	// Sync target file
	args[0] = "-f";
	args[1] = const_cast<char *>(targetpath1.c_str());
	ASSERT_INT_TRUE(ps->run_command("sync",0,1,args,2));

	// Edit the target file
	args[0] = const_cast<char *>(targetpath1.c_str());
	ASSERT_INT_TRUE(ps->run_command("edit",0,1,args,1));

	// Submit the target file
	args[0] = "-d";
	args[1] = "\"submit of test file\"";
	args[2] = const_cast<char *>(targetpath1.c_str());
	ASSERT_INT_TRUE(ps->run_command("submit",0,1,args,3));

	// Back to the original workspace
	ps->set_client("admin_space");

	// Do a parallel sync of the original workspace, causes overwrite error, should fail
	args[0] = "--parallel";
	args[1] = "threads=4,batch=8,batchsize=2000,min=9,minsize=2000";
	args[2] = const_cast<char *>(ppathall.c_str());
	int rv = ps->run_command("sync",0,1,args,3);
	ASSERT_INT_FALSE(rv);  // Expect failure

	ps->ReleaseConnection(0,0);

    return true;
}

bool TestP4BridgeServer::TestDefaultProgramNameAndVersion()
{
    P4ClientError* connectionError = NULL;
    // create a new server
    ps = new P4BridgeServer("localhost:6666", "admin", "", "admin_space");
    ASSERT_NOT_NULL(ps);

    // connect and see if the api returned an error. 
    if( !ps->connected( &connectionError ) )
    {
        char buff[256];
		snprintf(buff, 255, "Connection error: %s", connectionError->Message);
        // Abort if the connect did not succeed
        ASSERT_FAIL(buff);
    }
    char* params[1];
    params[0] = "//depot/mycode/*";

	//p4base::DumpMemoryState("Before first command run");

	ASSERT_INT_TRUE(ps->run_command("files", 6, 1, params, 1))

	// should both be set to default
	ASSERT_NOT_NULL(ps->pProgramName);
	ASSERT_NOT_NULL(ps->pProgramVer);
	ASSERT_STRING_EQUAL(ps->pProgramName, "something Application");
	ASSERT_STRING_EQUAL(ps->pProgramVer, "1, 0, 0, 1");

	//p4base::DumpMemoryState("After first command run");

	DELETE_ARRAY(ps->pProgramName);
	DELETE_ARRAY(ps->pProgramVer);

	ps->ReleaseConnection(6, 10);

	// set both program name and version
	ps->pProgramName = CopyStr(testProgramName);
	ps->pProgramVer = CopyStr(testProgramVer);

	ASSERT_INT_TRUE(ps->run_command("files", 7, 1, params, 1))

	// should both be set to supplied strings
	ASSERT_STRING_EQUAL(ps->pProgramName, testProgramName);
	ASSERT_STRING_EQUAL(ps->pProgramVer, testProgramVer);
	//p4base::DumpMemoryState("After first command run");

	DELETE_ARRAY(ps->pProgramName);
	DELETE_ARRAY(ps->pProgramVer);

	ps->ReleaseConnection(7, 10);

	// set program name but not version
	ps->pProgramName = CopyStr(testProgramName);
	ps->pProgramVer = NULL;

	ASSERT_INT_TRUE(ps->run_command("files", 8, 1, params, 1))

	// only name should  be set to supplied strings
	ASSERT_STRING_EQUAL(ps->pProgramName, testProgramName);
	ASSERT_STRING_NOT_EQUAL(ps->pProgramVer, testProgramVer);
	ASSERT_STRING_EQUAL(ps->pProgramVer, "1, 0, 0, 1");

	DELETE_ARRAY(ps->pProgramName);
	DELETE_ARRAY(ps->pProgramVer);

	ps->ReleaseConnection(8, 10);

	// set program version but not name
	ps->pProgramName = NULL;
	ps->pProgramVer = CopyStr(testProgramVer);

	ASSERT_INT_TRUE(ps->run_command("files", 7, 1, params, 1))

	// only version should  be set to supplied strings
	ASSERT_STRING_NOT_EQUAL(ps->pProgramName, testProgramName);
	ASSERT_STRING_EQUAL(ps->pProgramVer, testProgramVer);
	ASSERT_STRING_EQUAL(ps->pProgramName, "p4bridge-unit-test.exe");
	//p4base::DumpMemoryState("After first command run");

	DELETE_ARRAY(ps->pProgramName);
	DELETE_ARRAY(ps->pProgramVer);

	ps->ReleaseConnection(7, 10);

	int idleCnt = ps->FreeConnections(31);
    ASSERT_EQUAL(idleCnt, 0)
	//p4base::DumpMemoryState("After fourth FreeConnections");

    return true;
}

bool TestP4BridgeServer::TestGetTicketFile()
{
	char* tktfile = P4BridgeServer::GetTicketFile();

	ASSERT_NOT_NULL(tktfile);

	delete tktfile;

	return true;
}
